import { encoding, nativeChainIds, toChainId, } from "@wormhole-foundation/sdk-base";
import { canonicalAddress, serialize, toUniversal, universalAddress, } from "@wormhole-foundation/sdk-definitions";
import { EvmAddress, EvmPlatform, EvmUnsignedTransaction, addChainId, addFrom, } from "@wormhole-foundation/sdk-evm";
import "@wormhole-foundation/sdk-evm-core";
import { Ntt, } from "@wormhole-foundation/sdk-definitions-ntt";
import { Contract } from "ethers";
import { loadAbiVersion, } from "./bindings.js";
export class EvmNttWormholeTranceiver {
    manager;
    address;
    transceiver;
    constructor(manager, address, abiBindings) {
        this.manager = manager;
        this.address = address;
        this.transceiver = abiBindings.NttTransceiver.connect(address, manager.provider);
    }
    async getTransceiverType() {
        const contract = new Contract(this.address, ["function getTransceiverType() public view returns (string)"], this.manager.provider);
        try {
            const transceiverType = await contract
                .getFunction("getTransceiverType")
                .staticCall();
            if (!transceiverType) {
                throw new Error("getTransceiverType not found");
            }
            return Buffer.from(transceiverType).filter((x) => x !== 0).toString().trim();
        }
        catch (e) {
            console.error(e); // TODO: only when the method is not available should we return wormhole
            // NOTE: if the transceiver doesn't have a getTransceiverType method, it
            // means it's an old one that didn't have this function exposed.
            // Only wormhole transceivers were deployed using the old version
            return "wormhole";
        }
    }
    getAddress() {
        return {
            chain: this.manager.chain,
            address: toUniversal(this.manager.chain, this.address),
        };
    }
    encodeFlags(flags) {
        return new Uint8Array([flags.skipRelay ? 1 : 0]);
    }
    async *setPeer(peer) {
        const tx = await this.transceiver.setWormholePeer.populateTransaction(toChainId(peer.chain), universalAddress(peer));
        yield this.manager.createUnsignedTx(tx, "WormholeTransceiver.registerPeer");
    }
    async getPauser() {
        const pauser = await this.transceiver.pauser();
        return new EvmAddress(pauser);
    }
    async *setPauser(pauser) {
        const canonicalPauser = canonicalAddress({
            chain: this.manager.chain,
            address: pauser,
        });
        const tx = await this.transceiver.transferPauserCapability.populateTransaction(canonicalPauser);
        yield this.manager.createUnsignedTx(tx, "WormholeTransceiver.setPauser");
    }
    async getPeer(chain) {
        const peer = await this.transceiver.getWormholePeer(toChainId(chain));
        const peerAddress = encoding.hex.decode(peer);
        const zeroAddress = new Uint8Array(32);
        if (encoding.bytes.equals(zeroAddress, peerAddress)) {
            return null;
        }
        return {
            chain: chain,
            address: toUniversal(chain, peerAddress),
        };
    }
    async isEvmChain(chain) {
        return await this.transceiver.isWormholeEvmChain(toChainId(chain));
    }
    async *setIsEvmChain(chain, isEvm) {
        const tx = await this.transceiver.setIsWormholeEvmChain.populateTransaction(toChainId(chain), isEvm);
        yield this.manager.createUnsignedTx(tx, "WormholeTransceiver.setIsEvmChain");
    }
    async *receive(attestation) {
        const tx = await this.transceiver.receiveMessage.populateTransaction(serialize(attestation));
        yield this.manager.createUnsignedTx(tx, "WormholeTransceiver.receiveMessage");
    }
    async isRelayingAvailable(destination) {
        // TODO: how to handle executor?
        const [wh, special] = await Promise.all([
            this.isWormholeRelayingEnabled(destination),
            this.isSpecialRelayingEnabled(destination)
        ]);
        return wh || special;
    }
    async isWormholeRelayingEnabled(destChain) {
        return await this.transceiver.isWormholeRelayingEnabled(toChainId(destChain));
    }
    async *setIsWormholeRelayingEnabled(destChain, enabled) {
        const tx = await this.transceiver.setIsWormholeRelayingEnabled.populateTransaction(toChainId(destChain), enabled);
        yield this.manager.createUnsignedTx(tx, "WormholeTransceiver.setWormholeRelayingEnabled");
    }
    async isSpecialRelayingEnabled(destChain) {
        return await this.transceiver.isSpecialRelayingEnabled(toChainId(destChain));
    }
    async *setIsSpecialRelayingEnabled(destChain, enabled) {
        const tx = await this.transceiver.setIsSpecialRelayingEnabled.populateTransaction(toChainId(destChain), enabled);
        yield this.manager.createUnsignedTx(tx, "WormholeTransceiver.setSpecialRelayingEnabled");
    }
}
export class EvmNtt {
    network;
    chain;
    provider;
    contracts;
    version;
    tokenAddress;
    chainId;
    manager;
    xcvrs;
    managerAddress;
    constructor(network, chain, provider, contracts, version = "1.0.0") {
        this.network = network;
        this.chain = chain;
        this.provider = provider;
        this.contracts = contracts;
        this.version = version;
        if (!contracts.ntt)
            throw new Error("No Ntt Contracts provided");
        this.chainId = nativeChainIds.networkChainToNativeChainId.get(network, chain);
        this.tokenAddress = contracts.ntt.token;
        this.managerAddress = contracts.ntt.manager;
        const abiBindings = loadAbiVersion(this.version);
        this.manager = abiBindings.NttManager.connect(contracts.ntt.manager, this.provider);
        this.xcvrs = {};
        Object.entries(contracts.ntt.transceiver).map(([transceiverType, transceiver]) => {
            // we currently only support wormhole transceivers
            if (!["wormhole", "paxos"].includes(transceiverType)) {
                throw new Error(`Unsupported transceiver type: ${transceiverType}`);
            }
            // Enable more Transceivers here
            this.xcvrs[transceiverType] =
                new EvmNttWormholeTranceiver(this, transceiver instanceof Object ? transceiver.address : transceiver, abiBindings);
        });
    }
    async getTransceiver(type) {
        // TODO: should we make an RPC call here, or just trust that the xcvrs are set up correctly?
        return this.xcvrs[type] || null;
    }
    async getTransceivers() {
        return this.xcvrs;
    }
    async isTransceiverRegistered(type) {
        const transceiver = this.xcvrs[type];
        if (!transceiver) {
            throw new Error(`Transceiver not found (${type})`);
        }
        const transceivers = await this.manager.getTransceivers();
        return transceivers.map((a) => new EvmAddress(a).unwrap()).includes(new EvmAddress(transceiver.getAddress().address).unwrap());
    }
    async *registerTransceiver(type) {
        const transceiver = this.xcvrs[type];
        if (!transceiver) {
            throw new Error(`Transceiver not found (${type})`);
        }
        const canonicalTransceiver = new EvmAddress(transceiver.getAddress().address).toString();
        const tx = await this.manager.setTransceiver.populateTransaction(canonicalTransceiver);
        yield this.createUnsignedTx(tx, "Ntt.registerTransceiver");
    }
    async getMode() {
        const mode = await this.manager.getMode();
        return mode === 0n ? "locking" : "burning";
    }
    async isPaused() {
        return await this.manager.isPaused();
    }
    async *pause() {
        const tx = await this.manager.pause.populateTransaction();
        yield this.createUnsignedTx(tx, "Ntt.pause");
    }
    async *unpause() {
        const tx = await this.manager.unpause.populateTransaction();
        yield this.createUnsignedTx(tx, "Ntt.unpause");
    }
    async getOwner() {
        return new EvmAddress(await this.manager.owner());
    }
    async getPauser() {
        return new EvmAddress(await this.manager.pauser());
    }
    async *setOwner(owner) {
        const canonicalOwner = new EvmAddress(owner).toString();
        const tx = await this.manager.transferOwnership.populateTransaction(canonicalOwner);
        yield this.createUnsignedTx(tx, "Ntt.setOwner");
    }
    async *setPauser(pauser) {
        const canonicalPauser = new EvmAddress(pauser).toString();
        const tx = await this.manager.transferPauserCapability.populateTransaction(canonicalPauser);
        yield this.createUnsignedTx(tx, "Ntt.setPauser");
    }
    async *setThreshold(threshold) {
        const tx = await this.manager.setThreshold.populateTransaction(threshold);
        yield this.createUnsignedTx(tx, "Ntt.setThreshold");
    }
    async getThreshold() {
        return Number(await this.manager.getThreshold());
    }
    async isRelayingAvailable(destination) {
        const enabled = await Promise.all(Object.values(this.xcvrs).map((x) => x.isRelayingAvailable(destination)));
        return enabled.filter((x) => x).length > 0;
    }
    async getIsExecuted(attestation) {
        const payload = attestation.payloadName === "WormholeTransfer"
            ? attestation.payload
            : attestation.payload["payload"];
        const isExecuted = await this.manager.isMessageExecuted(Ntt.messageDigest(attestation.emitterChain, payload["nttManagerPayload"]));
        if (!isExecuted)
            return false;
        // Also check that the transfer is not queued for it to be considered complete
        return !(await this.getIsTransferInboundQueued(attestation));
    }
    async getIsTransferInboundQueued(attestation) {
        const payload = attestation.payloadName === "WormholeTransfer"
            ? attestation.payload
            : attestation.payload["payload"];
        return ((await this.getInboundQueuedTransfer(attestation.emitterChain, payload["nttManagerPayload"])) !== null);
    }
    getIsApproved(attestation) {
        const payload = attestation.payloadName === "WormholeTransfer"
            ? attestation.payload
            : attestation.payload["payload"];
        return this.manager.isMessageApproved(Ntt.messageDigest(attestation.emitterChain, payload["nttManagerPayload"]));
    }
    async getTokenDecimals() {
        return await EvmPlatform.getDecimals(this.chain, this.provider, this.tokenAddress);
    }
    async getPeer(chain) {
        const peer = await this.manager.getPeer(toChainId(chain));
        const peerAddress = encoding.hex.decode(peer.peerAddress);
        const zeroAddress = new Uint8Array(32);
        if (encoding.bytes.equals(zeroAddress, peerAddress)) {
            return null;
        }
        return {
            address: { chain: chain, address: toUniversal(chain, peerAddress) },
            tokenDecimals: Number(peer.tokenDecimals),
            inboundLimit: await this.getInboundLimit(chain),
        };
    }
    static async fromRpc(provider, config) {
        const [network, chain] = await EvmPlatform.chainFromRpc(provider);
        const conf = config[chain];
        if (conf.network !== network)
            throw new Error(`Network mismatch: ${conf.network} != ${network}`);
        const version = await EvmNtt.getVersion(provider, conf.contracts);
        return new EvmNtt(network, chain, provider, conf.contracts, version);
    }
    encodeOptions(options) {
        const ixs = [];
        ixs.push({
            index: 0,
            payload: this.xcvrs["wormhole"].encodeFlags({ skipRelay: !options.automatic }),
        });
        return ixs;
    }
    static async getVersion(provider, contracts) {
        const contract = new Contract(contracts.ntt.manager, ["function NTT_MANAGER_VERSION() public view returns (string)"], provider);
        try {
            const abiVersion = await contract
                .getFunction("NTT_MANAGER_VERSION")
                .staticCall();
            if (!abiVersion) {
                throw new Error("NTT_MANAGER_VERSION not found");
            }
            return abiVersion;
        }
        catch (e) {
            console.error(`Failed to get NTT_MANAGER_VERSION from contract ${contracts.ntt?.manager}`);
            throw e;
        }
    }
    async getCustodyAddress() {
        return this.managerAddress;
    }
    async quoteDeliveryPrice(dstChain, options) {
        const [, totalPrice] = await this.manager.quoteDeliveryPrice(toChainId(dstChain), Ntt.encodeTransceiverInstructions(this.encodeOptions(options)));
        return totalPrice;
    }
    async *setPeer(peer, tokenDecimals, inboundLimit) {
        const tx = await this.manager.setPeer.populateTransaction(toChainId(peer.chain), universalAddress(peer), tokenDecimals, inboundLimit);
        yield this.createUnsignedTx(tx, "Ntt.setPeer");
    }
    async *setWormholeTransceiverPeer(peer) {
        yield* this.setTransceiverPeer("wormhole", peer);
    }
    async *setTransceiverPeer(type, peer) {
        if (!this.xcvrs[type]) {
            throw new Error(`Transceiver not found (${type})`);
        }
        yield* this.xcvrs[type].setPeer(peer);
    }
    async *transfer(sender, amount, destination, options) {
        const senderAddress = new EvmAddress(sender).toString();
        // Note: these flags are indexed by transceiver index
        const totalPrice = await this.quoteDeliveryPrice(destination.chain, options);
        //TODO check for ERC-2612 (permit) support on token?
        const tokenContract = EvmPlatform.getTokenImplementation(this.provider, this.tokenAddress);
        const allowance = await tokenContract.allowance(senderAddress, this.managerAddress);
        if (allowance < amount) {
            const txReq = await tokenContract.approve.populateTransaction(this.managerAddress, amount);
            yield this.createUnsignedTx(addFrom(txReq, senderAddress), "Ntt.Approve");
        }
        const receiver = universalAddress(destination);
        const txReq = await this.manager
            .getFunction("transfer(uint256,uint16,bytes32,bytes32,bool,bytes)")
            .populateTransaction(amount, toChainId(destination.chain), receiver, receiver, options.queue, Ntt.encodeTransceiverInstructions(this.encodeOptions(options)), { value: totalPrice });
        yield this.createUnsignedTx(addFrom(txReq, senderAddress), "Ntt.transfer");
    }
    async *redeem(attestations) {
        for (const [transceiverType, transceiver] of Object.entries(this.xcvrs)) {
            const attestation = attestations[transceiverType];
            if (attestation?.payloadName !== "WormholeTransfer") {
                // TODO: support standard relayer attestations
                // which must be submitted to the delivery provider
                throw new Error("Invalid attestation type for redeem");
            }
            yield* transceiver.receive(attestation);
        }
    }
    async getCurrentOutboundCapacity() {
        return await this.manager.getCurrentOutboundCapacity();
    }
    async getOutboundLimit() {
        const encoded = (await this.manager.getOutboundLimitParams()).limit;
        const trimmedAmount = decodeTrimmedAmount(encoded);
        const tokenDecimals = await this.getTokenDecimals();
        return untrim(trimmedAmount, tokenDecimals);
    }
    async *setOutboundLimit(limit) {
        const tx = await this.manager.setOutboundLimit.populateTransaction(limit);
        yield this.createUnsignedTx(tx, "Ntt.setOutboundLimit");
    }
    async getCurrentInboundCapacity(fromChain) {
        return await this.manager.getCurrentInboundCapacity(toChainId(fromChain));
    }
    async getInboundLimit(fromChain) {
        const encoded = (await this.manager.getInboundLimitParams(toChainId(fromChain))).limit;
        const trimmedAmount = decodeTrimmedAmount(encoded);
        const tokenDecimals = await this.getTokenDecimals();
        return untrim(trimmedAmount, tokenDecimals);
    }
    async *setInboundLimit(fromChain, limit) {
        const tx = await this.manager.setInboundLimit.populateTransaction(limit, toChainId(fromChain));
        yield this.createUnsignedTx(tx, "Ntt.setInboundLimit");
    }
    async getRateLimitDuration() {
        return await this.manager.rateLimitDuration();
    }
    async getInboundQueuedTransfer(fromChain, transceiverMessage) {
        const queuedTransfer = await this.manager.getInboundQueuedTransfer(Ntt.messageDigest(fromChain, transceiverMessage));
        if (queuedTransfer.txTimestamp > 0n) {
            const { recipient, amount, txTimestamp } = queuedTransfer;
            const duration = await this.getRateLimitDuration();
            return {
                recipient: new EvmAddress(recipient),
                amount: amount,
                rateLimitExpiryTimestamp: Number(txTimestamp + duration),
            };
        }
        return null;
    }
    async *completeInboundQueuedTransfer(fromChain, transceiverMessage, payer) {
        const tx = await this.manager.completeInboundQueuedTransfer.populateTransaction(Ntt.messageDigest(fromChain, transceiverMessage));
        yield this.createUnsignedTx(tx, "Ntt.completeInboundQueuedTransfer");
    }
    async verifyAddresses() {
        const local = {
            manager: this.managerAddress,
            token: this.tokenAddress,
            transceiver: Object.fromEntries(Object.entries(this.xcvrs).map(([type, xcvr]) => [type, { address: xcvr.getAddress().address.toString() }])),
            // TODO: what about the quoter?
        };
        const remote = {
            manager: this.managerAddress,
            token: await this.manager.token(),
            transceiver: {
                wormhole: {
                    address: (await this.manager.getTransceivers())[0], // TODO: make this more generic
                }
            },
        };
        const deleteMatching = (a, b) => {
            if (!b || !a)
                return;
            for (const k in a) {
                if (typeof a[k] === "object") {
                    deleteMatching(a[k], b[k]);
                    if (Object.keys(a[k]).length === 0)
                        delete a[k];
                }
                else if (a[k] === b[k]) {
                    delete a[k];
                }
            }
        };
        deleteMatching(remote, local);
        return Object.keys(remote).length > 0 ? remote : null;
    }
    createUnsignedTx(txReq, description, parallelizable = false) {
        return new EvmUnsignedTransaction(addChainId(txReq, this.chainId), this.network, this.chain, description, parallelizable);
    }
}
function decodeTrimmedAmount(encoded) {
    const decimals = Number(encoded & 0xffn);
    const amount = encoded >> 8n;
    return {
        amount,
        decimals,
    };
}
function untrim(trimmed, toDecimals) {
    const { amount, decimals: fromDecimals } = trimmed;
    return scale(amount, fromDecimals, toDecimals);
}
function scale(amount, fromDecimals, toDecimals) {
    if (fromDecimals == toDecimals) {
        return amount;
    }
    if (fromDecimals > toDecimals) {
        return amount / 10n ** BigInt(fromDecimals - toDecimals);
    }
    else {
        return amount * 10n ** BigInt(toDecimals - fromDecimals);
    }
}
//# sourceMappingURL=ntt.js.map